% Euler's method of numerical integration for ordinary differential
% equations

clear all; clc

% Set simulation parameters & initialize system
dur = 1;  % duration of simulated proces in s
Dt = 0.01;  % time step for simulation
x = 1;  % initial state
gt_step = ceil(dur/(Dt*1000));  % graphical time step, ensures <= 1000 pts per plot 
DATA = zeros(2, floor(dur/gt_step));  % allocate memory
step = 0;  % # time steps since simulation began
n = 0;  % index of graph pts

% Run time loop
for t = 0:Dt:dur, step = step + 1;
  if mod(step, gt_step) == 0, n = n + 1;
    DATA(:, n) = [t; x];  % record data for plotting
  end
  x_vel = x; x = x + Dt*x_vel;  % x_vel is dx/dt
end  % for t
DATA = DATA(:, 1:n);

% Plot
figure(1), set(figure(1), 'Name', 'Euler Integration', 'NumberTitle', 'off');
plot(DATA(1, :), DATA(2, :), 'b'), ylabel('x'), xlabel('Time'), 
set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
