% A set of linear cells learns a linear function by the least mean square
% algorithm.

clear all; clc

% Initialize
nx = 8; ny = 2; ns = 30*nx;  % # elements in input vector x, output vector y, sample
x_sd = 1;  % standard deviation of each element of x 
W_STAR = (rand(ny, nx) - 0.5)/nx;  % desired weight matrix
W = (rand(ny, nx) - 0.5)/nx;  % initial weight matrix
y_sum = 0; y_sqr_sum = 0;
eta = 1/(nx + 2);  % learning-rate constant: 1/(nx + 2) optimal given indep Gaussian x(i)'s with SD 1
gt_step = ceil(ns/1000);  % graphical time step, ensures <= 1000 pts per plot
DATA = zeros(3, nx + floor(ns/gt_step));  % allocate more than enough memory
n = 0;  % index of graph points

% Train
for s = 1:ns
    
  % Set input and desired output
  x = x_sd*randn(nx, 1); y = W_STAR*x;

  % Process y so that later we can compute normalized squared error, NSE
  y_sum = y_sum + y; y_sqr_sum = y_sqr_sum + y'*y;
    
  % Compute network's output and error
  y_est = W*x;  % learner's estimate of y
  y_err = y_est - y;  % model error 
  
  % Adjust weights
  W = W - eta*y_err*x';
  WME = W - W_STAR;  % error in the weight matrix 
  WSE = sum(sum(WME.*WME));  % squared error of weights
  
  % Record data for plotting
  if (s <= nx) || (mod(s, gt_step) == 0), n = n + 1;
    DATA(:, n) = [s; y_err'*y_err; WSE];
  end

end  % for s
DATA = DATA(:, 1:n);

% Compute statistics of y
y_avg = y_sum/ns; y_var = y_sqr_sum/ns - y_avg'*y_avg;
DATA(2, :) = DATA(2, :)/y_var;  % normalized squared error, NSE

% Plot
figure(1), set(gcf, 'Name', 'LMS', 'NumberTitle', 'off');
subplot(2, 1, 1), plot(DATA(1, :), DATA(2, :), 'r'), ylabel('NSE')
xlim([DATA(1, 1) DATA(1, n)]), ylim([0 0.1])
line([1 gt_step*n], [0.01 0.01], 'LineStyle', '--', 'Color', 'k')
set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
subplot(2, 1, 2), plot(DATA(1, :), DATA(3, :), 'r'), ylabel('WSE'), xlabel('Example')
yl = ylim; ylim([0 yl(2)])
set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
