% A set of linear cells learns a linear operator by recursive least squares.

clear all; clc

% Initialize
nx = 8; ny = 2; ns = 10*nx;  % # elements in input vector x, output y, sample
W_STAR = (rand(ny, nx) - 0.5)/nx;  % desired weight matrix
W = (rand(ny, nx) - 0.5)/nx;  % initial weight matrix
P = 1e10*eye(nx);  % soft initialization
y_sum = 0; y_sqr_sum = 0;
gt_step = ceil(ns/1000);    % graph time step (for plots vs time, below)
DATA = zeros(3, nx + floor(ns/gt_step)); 
n = 0;  % index of graph point

% Train
for s = 1:ns
    
  % Set input and desired output
  x = randn(nx, 1); y = W_STAR*x;
 
  % Process y so that later we can compute normalized squared error, NSE
  y_sum = y_sum + y; y_sqr_sum = y_sqr_sum + y'*y;
  
  % Compute network's output and error
  y_est = W*x; y_err = y_est - y;
    
  % Adjust weights
  Px = P*x; k = Px'/(1 + x'*Px); P = P - Px*k; W = W - y_err*k;
  WME = W - W_STAR;  % error in the weight matrix 
  WSE = sum(sum(WME.*WME));  % squared error of weights
    
  % Record data for plotting
  if (s <= nx) || (mod(s, gt_step) == 0), n = n + 1;
    DATA(:, n) = [s; y_err'*y_err; WSE];
  end

end  % for s
DATA = DATA(:, 1:n);

% Compute statistics of y
y_avg = y_sum/ns; y_var = y_sqr_sum/ns - y_avg'*y_avg;
DATA(2, :) = DATA(2, :)/y_var;  % normalized squared error, NSE

% Plot
figure(1), set(gcf, 'Name', 'RLS', 'NumberTitle', 'off');
subplot(2, 1, 1), plot(DATA(1, :), DATA(2, :), 'r'), ylabel('NSE')
ylim([0 0.1])
line([DATA(1, 1) DATA(1, n)], [0.01 0.01], 'LineStyle', '--', 'Color', 'k')
set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
subplot(2, 1, 2), plot(DATA(1, :), DATA(3, :), 'r'), ylabel('WSE'), xlabel('Example')
xlim([DATA(1, 1) DATA(1, n)]), yl = ylim; ylim([0 yl(2)])
set(gca, 'TickDir', 'out', 'TickLength', [0.005 0.01])
